import platform, os, sys, stat, tempfile, re, subprocess
from browserstack.bserrors import BrowserStackLocalError
import gzip
import json

try:
    from urllib.request import urlopen, Request
except ImportError:
    from urllib2 import urlopen, Request

class LocalBinary:
  VERSION_REGEX = r"BrowserStack Local version \d+\.\d+"
  _version = None

  def __init__(self, key, error_object=None):
    self.key = key
    self.error_object = error_object
    is_64bits = sys.maxsize > 2**32
    self.is_windows = False
    osname = platform.system()
    source_url = self.fetch_source_url() + '/'

    if osname == 'Darwin':
      self.http_path = source_url + "BrowserStackLocal-darwin-x64"
    elif osname == 'Linux':
      if self.is_alpine():
        self.http_path = source_url + "BrowserStackLocal-alpine"
      elif platform.machine() == 'aarch64':
        self.http_path = source_url + "BrowserStackLocal-linux-arm64"
      else:
        if is_64bits:
          self.http_path = source_url + "BrowserStackLocal-linux-x64"
        else:
          self.http_path = source_url + "BrowserStackLocal-linux-ia32"
    else:
      self.is_windows = True
      self.http_path = source_url + "BrowserStackLocal.exe"

    self.ordered_paths = [
      os.path.join(os.path.expanduser('~'), '.browserstack'),
      os.getcwd(),
      tempfile.gettempdir()
    ]
    self.path_index = 0

  def fetch_source_url(self):
    url = "https://local.browserstack.com/binary/api/v1/endpoint"
    headers = {
      "Content-Type": "application/json",
      "Accept": "application/json",
      "User-Agent": '/'.join(('browserstack-local-python', LocalBinary._version))
    }
    data = {"auth_token": self.key}

    if self.error_object is not None:
      data["error_message"] = str(self.error_object)
      headers["X-Local-Fallback-Cloudflare"] = "true"

    req = Request(url, data=json.dumps(data).encode("utf-8"))
    for key, value in headers.items():
      req.add_header(key, value)

    try:
        with urlopen(req) as response:
          resp_bytes = response.read()
          resp_str = resp_bytes.decode('utf-8')
          resp_json = json.loads(resp_str)
          return resp_json["data"]["endpoint"]
    except Exception as e:
        raise BrowserStackLocalError('Error trying to fetch the source url for downloading the binary: {}'.format(e))

  @staticmethod
  def set_version(version):
    LocalBinary._version = version

  def is_alpine(self):
    response = subprocess.check_output(["grep", "-w", "NAME", "/etc/os-release"])
    return response.decode('utf-8').find('Alpine') > -1

  def __make_path(self, dest_path):
    try:
      if not os.path.exists(dest_path):
        os.makedirs(dest_path)
      return True
    except:
      return False

  def __available_dir(self):
    while self.path_index < len(self.ordered_paths):
      if self.__make_path(self.ordered_paths[self.path_index]):
        final_path = self.ordered_paths[self.path_index]
        self.path_index += 1
        return final_path
      else:
        self.path_index += 1
    raise BrowserStackLocalError('Error trying to download BrowserStack Local binary, exhausted user directories to download to.')

  def download(self, chunk_size=8192, progress_hook=None):
    headers = {
      'User-Agent': '/'.join(('browserstack-local-python', LocalBinary._version)),
      'Accept-Encoding': 'gzip, *',
    }

    if sys.version_info < (3, 2):
      # lack of support for gzip decoding for stream, response is expected to have a tell() method
      headers.pop('Accept-Encoding', None)

    response = urlopen(Request(self.http_path, headers=headers))
    try:
      total_size = int(response.info().get('Content-Length', '').strip() or '0')
    except:
      total_size = int(response.info().get_all('Content-Length')[0].strip() or '0')
    bytes_so_far = 0

    # Limits retries to the number of directories
    dest_parent_dir = self.__available_dir()
    dest_binary_name = 'BrowserStackLocal'
    if self.is_windows:
      dest_binary_name += '.exe'

    content_encoding = response.info().get('Content-Encoding', '')
    gzip_file = gzip.GzipFile(fileobj=response, mode='rb') if content_encoding.lower() == 'gzip' else None

    if os.getenv('BROWSERSTACK_LOCAL_DEBUG_GZIP') and gzip_file:
      print('using gzip in ' + headers['User-Agent'])

    def read_chunk(chunk_size):
      if gzip_file:
        return gzip_file.read(chunk_size)
      else:
        return response.read(chunk_size)

    with open(os.path.join(dest_parent_dir, dest_binary_name), 'wb') as local_file:
      while True:
        chunk = read_chunk(chunk_size)
        bytes_so_far += len(chunk)

        if not chunk:
          break

        if total_size > 0 and progress_hook:
          progress_hook(bytes_so_far, chunk_size, total_size)

        try:
          local_file.write(chunk)
        except:
          return self.download(chunk_size, progress_hook)
    
    if gzip_file:
      gzip_file.close()
    
    if callable(getattr(response, 'close', None)):
      response.close()

    final_path = os.path.join(dest_parent_dir, dest_binary_name)
    st = os.stat(final_path)
    os.chmod(final_path, st.st_mode | stat.S_IXUSR)
    return final_path

  def __verify_binary(self,path):
    try:
      binary_response = subprocess.check_output([path,"--version"]).decode("utf-8")
      return bool(re.match(LocalBinary.VERSION_REGEX, binary_response))
    except:
      return False

  def get_binary(self):
    dest_parent_dir = os.path.join(os.path.expanduser('~'), '.browserstack')
    if not os.path.exists(dest_parent_dir):
      os.makedirs(dest_parent_dir)
    binary_name = 'BrowserStackLocal.exe' if self.is_windows else 'BrowserStackLocal'
    bsfiles = [f for f in os.listdir(dest_parent_dir) if f == binary_name]
    
    if len(bsfiles) == 0:
      binary_path = self.download()
    else:
      binary_path = os.path.join(dest_parent_dir, bsfiles[0])

    valid_binary = self.__verify_binary(binary_path)
    if valid_binary:
      return binary_path
    else:
      binary_path = self.download()
      valid_binary = self.__verify_binary(binary_path)
      if valid_binary:
        return binary_path
      else:
        raise BrowserStackLocalError('BrowserStack Local binary is corrupt')




import subprocess, os, time, json, logging, re
import psutil

from browserstack.local_binary import LocalBinary
from browserstack.bserrors import BrowserStackLocalError

logger = logging.getLogger(__name__)
try:
    from importlib.metadata import version as package_version, PackageNotFoundError
except:
    import pkg_resources

class Local:
  def __init__(self, key=None, binary_path=None, **kwargs):
    self.key = os.environ['BROWSERSTACK_ACCESS_KEY'] if 'BROWSERSTACK_ACCESS_KEY' in os.environ else key
    self.options = kwargs
    self.local_logfile_path = os.path.join(os.getcwd(), 'local.log')
    LocalBinary.set_version(self.get_package_version())

  def __xstr(self, key, value):
    if key is None:
      return ['']
    if str(value).lower() == "true":
      return ['-' + key]
    elif str(value).lower() == "false":
      return ['']
    else:
      return ['-' + key, str(value)]

  def get_package_version(self):
    name = "browserstack-local"
    version = 'None'
    use_fallback = False
    try:
        temp = package_version
    except NameError: # Only catch if package_version is not defined(and not other errors)
        use_fallback = True

    if use_fallback:
        try:
            version = pkg_resources.get_distribution(name).version
        except pkg_resources.DistributionNotFound:
            version = 'None'
    else:
        try:
            version = package_version(name)
        except PackageNotFoundError:
            version = 'None'

    return version

  def _generate_cmd(self):
    cmd = [self.binary_path, '-d', 'start', '-logFile', self.local_logfile_path, "-k", self.key, '--source', 'python:' + self.get_package_version()]
    for o in self.options.keys():
      if self.options.get(o) is not None:
        cmd = cmd + self.__xstr(o, self.options.get(o))
    return cmd

  def _generate_stop_cmd(self):
    cmd = self._generate_cmd()
    cmd[2] = 'stop'
    return cmd

  def start(self, **kwargs):
    for k, v in kwargs.items():
        self.options[k] = v

    if 'key' in self.options:
      self.key = self.options['key']
      del self.options['key']

    if 'binarypath' in self.options:
      candidate = os.path.realpath(self.options['binarypath'])
      if not os.path.isfile(candidate):
        raise BrowserStackLocalError('binarypath does not point to a file')
      try:
        version_output = subprocess.check_output([candidate, '--version'], timeout=10).decode('utf-8')
      except (subprocess.SubprocessError, OSError) as e:
        raise BrowserStackLocalError('binarypath failed verification: {}'.format(e))
      if not re.match(LocalBinary.VERSION_REGEX, version_output):
        raise BrowserStackLocalError('binarypath failed verification')
      self.binary_path = candidate
      del self.options['binarypath']
    else:
      l = LocalBinary(self.key)
      try:
        self.binary_path = l.get_binary()
      except Exception as e:
        l = LocalBinary(self.key, e)
        self.binary_path = l.get_binary()

    if 'logfile' in self.options:
      self.local_logfile_path = self.options['logfile']
      del self.options['logfile']

    if "onlyCommand" in kwargs and kwargs["onlyCommand"]:
      return

    if 'source' in self.options:
      del self.options['source']

    logfile_dir = os.path.dirname(self.local_logfile_path)
    if logfile_dir:
        os.makedirs(logfile_dir, exist_ok=True)
    try:
        with open(self.local_logfile_path, 'w') as f:
            f.write('')
    except OSError as e:
        raise BrowserStackLocalError('Unable to open logfile: {}'.format(e))

    self.proc = subprocess.Popen(self._generate_cmd(), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    (out, err) = self.proc.communicate()

    try:
      if out:
        output_string = out.decode()
      else:
        output_string = err.decode()

      data = json.loads(output_string)

      if data['state'] != "connected":
        raise BrowserStackLocalError(data["message"]["message"])
      else:
        self.pid = data['pid']
    except ValueError:
      logger.error("BinaryOutputParseError: Raw String = '{}'".format(output_string) )
      raise BrowserStackLocalError('Error parsing JSON output from daemon. Raw String = "{}"'.format(output_string))

  def isRunning(self):
    return hasattr(self, 'pid') and psutil.pid_exists(self.pid)

  def stop(self):
    try:
      proc = subprocess.Popen(self._generate_stop_cmd(), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
      (out, err) = proc.communicate()
    except Exception as e:
      return

  def __enter__(self):
    self.start(**self.options)
    return self

  def __exit__(self, *args):
    self.stop()
